use std::ops::{Index, IndexMut, RangeFull};
use std::iter::IntoIterator;


#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub struct Stack<T, const S: usize> {
    data: [T; S],
    len: usize // we can experiment with using u8 some other time
}


impl<T, const S: usize> Stack<T, S> {
    pub fn push(&mut self, v: T) {
        self.data[self.len] = v;
        self.len += 1;
    }

    pub fn len(&self) -> usize {
        self.len
    }

    pub fn clear(&mut self) {
        self.len = 0;
    }

    pub fn last(&self) -> Option<&T> {
        if self.len == 0 {
            None
        }
        else {
            Some(&self.data[self.len - 1])
        }
    }

    pub fn iter(&self) -> impl Iterator<Item = &T> {
        self.data.iter().take(self.len)
    }

    pub const fn from_array(array: [T; S]) -> Self {
        Stack {
            data: array,
            len: S,
        }
    }

    pub fn into_inner(self) -> [T; S] {
        self.data
    }
}


impl<T, const S: usize> Stack<T, S>
where T: Copy + Default
{
    pub fn new() -> Self {
        Stack {
            data: [Default::default(); S],
            len: 0,
        }
    }

    pub fn move_slice(&mut self, dst: &mut Self, start: usize) {
        let slice_len = self.len - start;
        let src_slice = &mut self.data[start..self.len];
        let dst_slice = &mut dst.data[dst.len..(dst.len + slice_len)];
        dst_slice.copy_from_slice(src_slice);

        self.len -= slice_len;
        dst.len += slice_len;
    }

    pub fn move_slice_under(&mut self, dst: &mut Self, start: usize) {
        let slice_len = self.len - start;
        let src_slice = &mut self.data[start..self.len];
        
        dst.data.rotate_right(slice_len);
        let dst_slice = &mut dst.data[0..slice_len];

        dst_slice.copy_from_slice(src_slice);

        self.len -= slice_len;
        dst.len += slice_len;
    }

    // like above, except source and destination are the same, i.e. reordering the stack
    pub fn shift_slice_under(&mut self, start: usize) {
        for mut i in start..self.len {
            while i > 0 {
                self.data.swap(i, i -1);
                i -= 1;
            }
        }
    }
}


impl<T, const S: usize> Default for Stack<T, S>
where T: Copy + Default
{
    fn default() -> Self {
        Self::new()
    }
}


impl<T, const S: usize> Index<usize> for Stack<T, S> {
    type Output = T;
    fn index(&self, index: usize) -> &T {
        &self.data[index]
    }
}

impl<T, const S: usize> Index<RangeFull> for Stack<T, S> {
    type Output = [T];
    fn index(&self, _index: RangeFull) -> &[T] {
        &self.data[..self.len]
    }
}

impl<T, const S: usize> IndexMut<RangeFull> for Stack<T, S> {
    fn index_mut(&mut self, _index: RangeFull) -> &mut [T] {
        &mut self.data[..self.len]
    }
}


impl<I, T, const S: usize> From<I> for Stack<T, S> 
where 
    T: Copy + Default,
    I: IntoIterator<Item = T>
{
    fn from(src: I) -> Self {
        let mut res = Self::new();
        for (i, item) in src.into_iter().enumerate() {
            if i >= S {
                break;
            }
            res.push(item);
        }
        res
    }
}


#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_basic() {
        let mut stack: Stack<usize, 5> = Stack::new();
        stack.push(1);
        stack.push(2);
        stack.push(3);

        assert_eq!(stack.len(), 3);
        assert_eq!(stack[0], 1);
        assert_eq!(stack[1], 2);
        assert_eq!(stack[2], 3);

        assert_eq!(stack.last(), Some(&3));

        stack.clear();
        assert_eq!(stack.len(), 0);
    }

    #[test]
    fn test_move_slice() {
        let mut a: Stack<usize, 5> = Stack::new();
        let mut b: Stack<usize, 5> = Stack::new();

        a.push(1);
        a.push(2);
        a.push(3);
        b.push(9);
        b.push(8);

        a.move_slice(&mut b, 1);
        assert_eq!(b[2], 2);
        assert_eq!(b[3], 3);

        b.move_slice(&mut a, 1);
        assert_eq!(a[1], 8);
        assert_eq!(a[2], 2);
        assert_eq!(a[3], 3);

        a.move_slice(&mut b, 0);
        assert_eq!(a.len(), 0);
        assert_eq!(b[0], 9);
        assert_eq!(b.last(), Some(&3));
    }

    #[test]
    fn test_move_slice_under() {
        let mut a: Stack<usize, 5> = Stack::new();
        let mut b: Stack<usize, 5> = Stack::new();

        a.push(1);
        a.push(2);
        a.push(3);
        b.push(9);
        b.push(8);

        a.move_slice_under(&mut b, 1);
        assert_eq!(a.len(), 1);
        assert_eq!(a[0], 1);
        
        assert_eq!(b.len(), 4);
        assert_eq!(b[0], 2);
        assert_eq!(b[3], 8);

        b.move_slice_under(&mut a, 0);
        assert_eq!(b.len(), 0);
        assert_eq!(a[0], 2);
        assert_eq!(a[4], 1);
    }

    fn test_shift_slice_under() {
        let mut a: Stack<usize, 5> = Stack::from([1, 2, 3, 4, 5]);
        a.shift_slice_under(3);
        assert_eq!(a[0], 4);
        assert_eq!(a[1], 5);
        assert_eq!(a[2], 1);
        assert_eq!(a[3], 2);
        assert_eq!(a[4], 3);
    }

    #[test]
    fn test_from_iter() {
        let s = Stack::<_, 5>::from([1, 2, 3]);
        assert_eq!(s[0], 1);
        assert_eq!(s[2], 3);

        let s = Stack::<_, 2>::from([1, 2, 3]);
        assert_eq!(s.len(), 2);
        assert_eq!(s[0], 1);
        assert_eq!(s[1], 2);
    }

    #[test]
    fn test_iter() {
        let s = Stack::<_, 5>::from([1, 2, 3]);
        let mut it = s.iter();
        assert_eq!(it.next(), Some(&1));
        assert_eq!(it.next(), Some(&2));
        assert_eq!(it.next(), Some(&3));
        assert_eq!(it.next(), None);
    }

    #[test]
    fn test_from_array() {
        let s = Stack::from_array([1, 2, 3]);
        assert_eq!(s[0], 1);
        assert_eq!(s[1], 2);
        assert_eq!(s[2], 3);
        assert_eq!(s.len(), 3);
    }

    #[test]
    fn test_slice_index() {
        let mut s = Stack::<_, 5>::from([3, 4, 5]);
        assert_eq!(s[..], [3, 4, 5]);
        assert_eq!(&mut s[..], &mut [3, 4, 5]);
    }
}
